#pragma rtGlobals=1		// Use modern global access method.

#include "Neuralynx Analysis"

//	NOTE: This requires version 1.25 or later Neuralynx files which start with a 16,384 byte header.
//	Earlier files lack this header.
// 
//	See http://www.neuralynx.com/download/NeuralynxDataFileFormats.pdf for file format specifications.  
//	 
//	This supports .ntt and .ncs files.  A "Neuralynx" menu item is provided in the Analysis menu which provides panel access to the individual file loading and saving functions.  All of the loading functions 
//	are also accessible from the Data/Load Waves menu.  The types variable throughout these function refer to "ntt" or "ncs" (must not contain a leading period) files.  For .ntt files, the data will be loaded
//	as a data wave, a wave with a "_t" suffix (spike times) and a wave with a "_c" suffix (cluster assignments if SpikeSort3D has been used), and a wave with a "_h" suffix (the file header).  The .ntt data 
//	wave will have 3 dimensions; each spike gives one row.  There are 32 columns because each spike contains 32 samples per channel, and there are 4 layers, one for each channel.  The panel provides
// 	access to a viewer which allows for review of the loaded spike data.  There are also some rudimentary analysis functions such as spike-time-histograms and PSTHs if stimulus times are known.  
//	Primitive clustering is available via PCA but is not recommended since better tools like SpikeSort3D or MClust are available.  Many of the functions provided here do not apply to .ncs (continous) data, 
//	except for the loaders, since there isn't much more to do with these that one can't do on one's own.  

strconstant NlxFolder="root:Packages:Neuralynx:"

// Information that proceeds each chunk of data in .ncs files.   
constant NCSInfoLength=20 // Total bytes per record (8 timestamp, 4 channel number, 4 sampling frequency, 4 num valid samples).  
constant NCSDataSize=2 // Bytes per point (int16).  
constant NCSDataLength=1024 //  Bytes per record. (int16 x 512).  
Structure NlxNCSInfo
	uint32 TimeStamp_low
	uint32 TimeStamp_high
	uint32 ChannelNumber
	uint32 SampleFreq
	uint32 NumValidSamples
EndStructure

// Information that proceeds each chunk of data in .ntt files.   
constant NTTInfoLength=48 // Total bytes per record (8 timestamp, 4 acquisition entity, 4 for classified cell number, 4x8 for spike feature)	.
constant NTTDataSize=2 // Bytes per point (int16).  
constant NTTDataLength=256 // Bytes per record. (int16 x 32 x 4).  
Structure NlxNTTInfo
	uint32 TimeStamp_low
	uint32 TimeStamp_high
	uint32 SpikeAcqNumber
	uint32 ClassCellNumber
	uint32 FeatureParams[8]
EndStructure

// Things we want to read out of each file's text header.  
static constant headerSize = 16384			// Size of header in version 1.25 and later files
Structure NlxHeader
	double SampleFreq 
	double bitVolts
EndStructure

static constant timestamp_scale=1000000 // The timestamp units are in microseconds.  

// Add an item to Igor's Load Waves submenu (Data menu)
Menu "Load Waves"
	SubMenu "Load Neuralynx"
		"Load Neuralynx Binary Continuous File...", LoadNlxBinaryFile("ncs","")
		"Load Neuralynx Binary Tetrode File...", LoadNlxBinaryFile("ntt","")
		"Load All Neuralynx Binary Continuous Files In Folder...", LoadAllNlxBinaryFiles("ncs","")
		"Load All Neuralynx Binary Tetrode Files In Folder...", LoadAllNlxBinaryFiles("ntt","")
		//"Edit Neuralynx Loader Settings", EditNlxDefaultSettings()
	End
End

// For editing the default settings of the Neuralynx loader.  Not yet implemented.  To use non-default settings, use optional variables available in the functions.  
//Function EditNlxDefaultSettings()
//End
//
//static Function GetNlxDefaultSettings(type,channelBaseNameDefault, createTableDefault, createGraphDefault)
//	String type
//	String &channelBaseNameDefault				// Output
//	Variable &createTableDefault						// Output
//	Variable &createGraphDefault					// Output
//	
//	type=type[1,strlen(type)-1]
//	strswitch(type)
//		case "ncs":
//			String default_name="Continuous"
//			break
//		case "ntt":
//			default_name="Tetrode"
//			break
//	endswitch
//	
//	String Nlx_folder="root:Packages:Nlx:"
//	channelBaseNameDefault = StrVarOrDefault(Nlx_folder+type+":channelBaseName", default_name)
//	createTableDefault = NumVarOrDefault(Nlx_folder+type+":createTable", 1)
//	createGraphDefault = NumVarOrDefault(Nlx_folder+type+":createGraph", 1)
//End
//
//static Function SetNlxDefaultSettings(type,channelBaseNameDefault, createTableDefault, createGraphDefault)
//	String type, channelBaseNameDefault
//	Variable createTableDefault, createGraphDefault
//	
//	String curr_folder=GetDataFolder(1)
//	NewDataFolder/O/S root:Packages
//	NewDataFolder/O/S Nlx
//	type=type[1,strlen(type)-1]
//	NewDataFolder /O/S $type
//	String/G channelBaseName = channelBaseNameDefault
//	Variable/G createTable = createTableDefault
//	Variable/G createGraph = createGraphDefault
//	SetDataFolder $curr_folder
//End

// Loads the header of a neuralynx file.  
Function LoadNlxHeader(refNum, type, NlxHeader)
	Variable refNum
	String type
	Struct NlxHeader &NlxHeader		// Outputs
	
	String file_open_line,file_close_line,sampleFreq_line,bitVolts_line
	Variable bitVolts,sampleFreq
	FSetPos refNum, 0
	
	FSkipLines(refNum,2)
	FReadLine refNum,file_open_line // Currently unused.  
	FReadLine refNum,file_close_line // Currently unused.  
	
	FSkipLines(refNum,8)
	FReadLine refNum,sampleFreq_line
	sscanf sampleFreq_line,"-SamplingFrequency %f",sampleFreq
	NlxHeader.sampleFreq=sampleFreq
	
	FSkipLines(refNum,1)
	FReadLine refNum,bitVolts_line
	strswitch(type)
		case "ncs":
			sscanf bitVolts_line,"-ADBitVolts %f",bitVolts
			break
		case "ntt":
			sscanf bitVolts_line,"-ADBitVolts %f",bitVolts // Even though tetrodes could have 4 settings, they are likely to be the same.  
			break
	endswitch
	NlxHeader.bitVolts=bitVolts
End

// Loads the data (everything aftet the header) from a Neuralynx file.  
static Function /S LoadNlxData(refNum, type, baseName, NlxHeader)
	Variable refNum
	String type,baseName
	Struct NlxHeader &NlxHeader
		
	FSetPos refNum, headerSize													// Go to the end of the header (start of the data).  
	FStatus refNum
	Variable i,j,bytes=V_logEOF-V_filePos
	strswitch(type)
		case "ncs":
			Variable numRecords=bytes/(NCSInfoLength+NCSDataLength) // Number of 512 point frames of continuous data plus associated meta-information.  
			Struct NlxNCSInfo NCSInfo
			FBinRead/B=3 refNum, NCSInfo						// Read the info for the first frame of this continuous event.  
			Variable begin_time=(NCSInfo.timestamp_low+(2^32)*NCSInfo.timestamp_high)/timestamp_scale // Only useful if there are no interruptions in acquisition.  
			Variable chan=NCSInfo.ChannelNumber
			Variable wave_delta=1/NCSInfo.SampleFreq
			break
		case "ntt":
			numRecords = bytes/(NTTInfoLength+NTTDataLength) // Number of distinct tetrode records.  
			Struct NlxNTTInfo NTTInfo
			FBinRead/B=3 refNum, NTTInfo						// Read the info for the first frame of this tetrode event.  
			begin_time=0 									// Each event has its own timestamp.  
			chan=NTTInfo.SpikeAcqNumber
			wave_delta=1/NlxHeader.SampleFreq // Sampling frequency given in the header but not in the frame info for tetrode data.  
			break
	endswitch

	if(!strlen(baseName))
		baseName=type
	endif
	String wave_name=baseName+num2str(chan)
	Make/O/n=(bytes/numRecords,numRecords) /B/U $(wave_name); Wave Data=$wave_name // uint8 wave (byte wave).  
	FSetPos refNum,headerSize 								// Go back to the beginning of the data.  
	FBinRead/B=3 refNum, Data								// Read the rest of the file.  	
	
	// Now sort the wheat from the chaff (clean up the data wave to remove bytes containing metainformation, and resize/redimension it.)  
	NlxIgorizeData(Data,type,numRecords)
	
	strswitch(type)
		case "ncs":
			SetScale/P x, begin_time, wave_delta, "s", Data
			break
		case "ntt":
			break
	endswitch
	
	return wave_name
End

// Takes Igor waves of data and times, created by LoadNlxData, and saves it back into a Neuralynx file.  Assumes there will be already be an intact header in the file that will be overwritten.  
// Use this if you have done some processing on the data in Igor, like removing bogus spikes or extracting a subset of the data, and now want to re-export it for analysis with Neuralynx tools.  
static Function /S SaveNlxData(refNum, type, DataWave, TimeWave)
	Variable refNum
	String type
	Wave DataWave, TimeWave
	
	strswitch(type)
		case "ncs":
			break
		case "ntt":
			Variable numRecords = numpnts(TimeWave) // Number of distinct tetrode records.  
	endswitch
	
	String wave_name="NxFormatted_"+type
	Make/O/n=0 /B/U $(wave_name) /WAVE=NxFormatted // uint8 wave (byte wave).  
	NlxDeIgorizeData(DataWave,TimeWave,type,NxFormatted)
	FStatus refNum
	FSetPos refNum,headerSize 								// Go back to the beginning of the data.  
	print numpnts(NxFormatted)
	FBinWrite/B=3 refNum, NxFormatted						// Write the new data.  
End

// All the Neuralynx data after the header consists of a series of frames.  Each frame contains data and metadata.  This converts the old metadata+data (Raw) into just the data (Raw) as well 
// as a wave of cluster numbers (_c suffix) and a wave of times (_t suffix).  
Function NlxIgorizeData(Raw,type,numRecords)
	Wave Raw
	String type
	Variable numRecords
	
	String wave_name=NameOfWave(Raw)	
	Variable i
	//Make /o/n=(numRecords)/I NumValidSamples // Always 512 according to Brian at Nlx.  
	strswitch(type)
		case "ncs":
			break
		case "ntt":
			Make /o/D/U/n=(numRecords) $(wave_name+"_t") /WAVE=Times
			Make /o/n=(numRecords) $(wave_name+"_c") /WAVE=Clusters 
			break
	endswitch
	for(i=0;i<numRecords;i+=1)
		Variable timestamp=0
		strswitch(type)
			case "ncs":
				Struct NlxNCSInfo NCSInfo
				StructGet /B=3 NCSInfo, Raw[i]
				timestamp=NCSInfo.timestamp_low+(2^32)*NCSInfo.timestamp_high
				//NumValidSamples[i]=NCSInfo.NumValidSamples
				break
			case "ntt":
				Struct NlxNTTInfo NTTInfo
				StructGet /B=3 NTTInfo, Raw[i]
				timestamp=NTTInfo.timestamp_low+(2^32)*NTTInfo.timestamp_high
				Times[i]=timestamp/timestamp_scale
				Clusters[i]=NTTInfo.ClassCellNumber
				break
		endswitch
		
	endfor
	strswitch(type)
			case "ncs":
				DeletePoints /M=0 0,NCSInfoLength,Raw // Delete the metainformation.  
				Redimension /n=(numpnts(Raw)/NCSDataSize)/E=1/W Raw // Convert into one long wave with the approriate point size.  
				break
			case "ntt":
				DeletePoints /M=0 0,NTTInfoLength,Raw // Delete the metainformation.  
				Redimension /n=(numpnts(Raw)/NTTDataSize)/E=1/W Raw // Convert into one long wave with the appropriate point size.    
				Duplicate /O Raw, NTTTemp
				Redimension /n=(32,numRecords,4) Raw // Tetrode number (not sample number) is the first index, so redimensioning in one step doesn't work.  
				Raw[][][]=NTTTemp[(p*4)+(128*q)+r] // Maybe use MatrixOp with WaveMap to improve speed?  No, it only supports 2 dimensional waves.  
				KillWaves /Z NTTTemp
				break
	endswitch
	// Raw is now clean!  
End

// The opposite of NlxIgorizeData.  This puts the Data and the Times (but not the clusters) back into a Neuralynx frame format for use with re-exporting Neuralynx data with SaveNeuralynxData.  
Function NlxDeIgorizeData(DataWave,TimeWave,type,NxFormatted)
	Wave DataWave,TimeWave,NxFormatted
	String type
	
	Variable i,j,k,numRecords=numpnts(TimeWave)
	Variable numRows=dimsize(NxFormatted,0)
	Struct NlxHeader NlxHeader
	Wave Header=root:Packages:Nlx:ntt:$(NameOfWave(DataWave)+"_h")
	StructGet NlxHeader Header
	//Make/O/n=(bytes/numRecords,numRecords) /B/U $(wave_name) /WAVE=NxFormatted // uint8 wave (byte wave).
	//Make /o/n=(numRecords)/I NumValidSamples // Always 512 according to Brian at Nlx.  
	for(i=0;i<numRecords;i+=1)
		Variable timestamp=0
		strswitch(type)
			case "ncs":
				break
			case "ntt":
				Struct NlxNTTInfo NTTInfo
				NTTInfo.timestamp_low=mod(TimeWave[i],2^32)
				NTTInfo.timestamp_high=floor(TimeWave[i]/(2^32))
//				NTTInfo.SpikeAcqNumber=0 // Ignore this since it doesn't matter.  
//				NTTInfo.ClassCellNumber=0 // Ignore this since it doesn't matter.  
//				for(j=0;j<8;j+=1)
//					NTTInfo.FeatureParams[j]=0// Ignore this since it doesn't matter.  
//				endfor
				StructPut /B=3 NTTInfo, NxFormatted[i]
				break
		endswitch
	endfor
	strswitch(type)
			case "ncs":
				break
			case "ntt":  
				Duplicate /o DataWave,NTTTemp
				Redimension /n=(numpnts(DataWave))/W NTTTemp
				NTTTemp=DataWave[floor(mod(p,128)/4)][floor(p/128)][mod(p,4)]/NlxHeader.bitVolts
				Redimension /n=(NTTDataSize*32*4,numRecords)/E=1/B/U NTTTemp
				print numRecords
				Redimension /n=(NTTInfoLength+32*4*NTTDataSize,numRecords) NxFormatted
				NxFormatted[NTTInfoLength,][]=NTTTemp[p-NTTInfoLength][q] // Maybe use MatrixOp with WaveMap to improve speed?    
				break
	endswitch
	// Raw is now clean!  
End

// Load a Neuralynx file.  
Function /S LoadNlxBinaryFile(type,fileName[,pathName,baseName,createTable,createGraph,downSamp])
	String type					// File type: continuous "ncs" or tetrode "ntt".   
	String fileName				// file name, partial path and file name, or full path or "" for dialog.
	String pathName				// Igor symbolic path name or "" for dialog.
	String baseName				// Base name for the wave for this channel .  
	Variable createTable			// 1 = Yes, 2 = No, 0 for dialog.
	Variable createGraph			// 1 = Yes, 2 = No, 0 for dialog.
	Variable downSamp
	
	strswitch(type)
		case "ncs":
			break
		case "ntt":
			break
		default:
			DoAlert 0, "Unknown format: "+type
			return num2str(-100) // Exit now so we don't have to deal with the default case again.  
			break
	endswitch
	
	if(ParamIsDefault(pathName))
		PathInfo NlxPath
		if(V_flag)
			pathName="NlxPath"
		else
			pathName="home"
		endif
	endif

	if(ParamIsDefault(baseName))
		baseName=StrVarOrDefault(NlxFolder+"baseNameDefault",type)
	endif
	if(ParamIsDefault(createTable))
		 createTable=NumVarOrDefault(NlxFolder+"createTableDefault",0)
	endif
	if(ParamIsDefault(createGraph))
		createGraph=NumVarOrDefault(NlxFolder+"createGraphDefault",0)
	endif
	
	String message = "Select a Nlx binary ."+type+" file"
	Variable refNum
	Open/R/Z=2/P=$pathName/M=message/T=("."+type) refNum as fileName+"."+type
	
	// Save outputs from Open in a safe place.
	if (V_flag != 0)
		return num2str(V_flag)			// -1 means user canceled.
	endif
	String fullPath = S_fileName
	String path=RemoveListItem(ItemsInList(fullPath,":")-1,fullPath,":")
	NewPath /O/Q/Z NlxPath path
	
	printf "Loading Nlx binary "+type+" data from \"%s\"\r", fullPath
	
	Struct NlxHeader NlxHeader
	LoadNlxHeader(refNum,type,NlxHeader)
	
	FStatus refNum
	if(V_logEOF<=16384) // Has only the header or even less.  
		print fullPath+" has no data"
		return num2str(-200)
	else
		String loadedName=LoadNlxData(refNum, type, baseName, NlxHeader)	// Load the sample data
	endif
	
	Make /o $(loadedName+"_h") /WAVE=Header
	StructPut NlxHeader Header // Save the header file in case we need to export a modified Nlx file.  

	Wave NlxData = $loadedName									// Create reference to wave created by LoadNlxData.
	
	if(downSamp>1)
		Wave Loaded=$loadedName
		if(numpnts(Loaded)>100000000)
			Variable width=dimdelta(Loaded,0)*downSamp
			//Variable left=x2pnt(leftx(Loaded), right=left+(rightx(Loaded)-leftx(Loaded))/downSamp
			Loaded[0,dimsize(Loaded,0)/downSamp]=mean(Loaded,x,x+width)
			Redimension /n=(dimsize(Loaded,0)/downSamp,-1) Loaded
			print numpnts(Loaded)
		else
			Resample /DOWN=(downSamp) Loaded
		endif
	endif
	Redimension/S NlxData												// Change to floating point so we can represent data in volts.
	
	NlxData*=NlxHeader.bitVolts
	SetScale d, 0, 0, "V", NlxData										// Note that the data units are volts
		
	Variable numSamples = numpnts(NlxData)
	Printf "Created wave %s. %d points.\r", loadedName, numSamples
	if (createTable)
		Edit NlxData.id
		ModifyTable format[1]=3,digits[1]=6										// Display time in suitable format.
	endif
	if (createGraph)
		strswitch(type)
			case "ncs":
				Display NlxData
				break
			case "ntt":
				CreateNlxViewer(NlxData,"Tetrode")
				break
		endswitch
	endif

	Close refNum

	return loadedName			
End

// Using Igor components, overwrite the data component of a Neuralynx file.  
Function SaveNlxBinaryFile(type,DataWave[,pathName,fileName])
	String type					// File type: continuous "ncs" or tetrode "ntt".   
	Wave DataWave				// The data wave to be written to a file.  
	String pathName				// Igor symbolic path name or "" for dialog.
	String fileName				// file name, partial path and file name, or full path or "" for dialog.
	
	strswitch(type)
		case "ncs":
			break
		case "ntt":
			break
		default:
			DoAlert 0, "Unknown format: "+type
			return -2 // Exit now so we don't have to deal with the default case again.  
			break
	endswitch
	
	Variable err
	
	if(ParamIsDefault(pathName))
		if(!exists("root:Packages:Nlx:defaultPath"))
			Do
				NewPath /O/M="Choose a default folder for Nlx files."/Q NlxPath
			While(V_flag)
			String /G root:Packages:Nlx:defaultPath="NlxPath"
		endif
		SVar /Z defaultPath=root:Packages:Nlx:defaultPath
		pathName=defaultPath
	endif

	// This puts up a dialog if the pathName and fileName do not specify a file.
	String message = "Select a Nlx binary "+type+" file to overwrite."
	Variable refNum
	if(ParamIsDefault(fileName))
		Open /P=$pathName/M=message/T=("."+type) refNum
	else
		Open/R/Z=2/P=$pathName/M=message/T=("."+type) refNum as fileName
	endif
	
	// Save outputs from Open in a safe place.
	err = V_Flag
	String fullPath = S_fileName

	if (err != 0)
		return err			// -1 means user canceled.
	endif
	
	Printf "Overwriting Nlx binary "+type+" data in \"%s\"\r", fullPath
	
	FStatus refNum
	if(V_logEOF<16384) // Does not have a full header.   
		print fullPath+" has no header"
		return 0
	else
		Wave TimeWave=$(GetWavesDataFolder(DataWave,2)+"_t")
		SaveNlxData(refNum, type, DataWave, TimeWave)	// Load the sample data
	endif

	Close refNum
	return 0			// Zero signifies no error.	
End

// Loads all the neuralynx files of a given type in a directory.  
Function LoadAllNlxBinaryFiles(type,baseName[,pathName,createTable,createGraph,downSamp])
	String type				// File type.  
	String pathName			// Name of an Igor symbolic folder created by Misc->NewPath. "" for dialog.
	String baseName	// Base name for the wave for this channel or "" for dialog.
	Variable createTable		// 1 = Yes, 2 = No, 0 for dialog.
	Variable createGraph		// 1 = Yes, 2 = No, 0 for dialog.
	Variable downSamp
	
	if(ParamIsDefault(pathName) || strlen(pathName) == 0)
		String message = "Select a directory containing Nlx "+type+" files"
		NewPath/O/Q/M=message NlxDataPath		// This displays a dialog in which you can select a folder
		if (V_flag != 0)
			return V_flag									// -1 means user canceled
		endif
		pathName = "NlxDataPath"
	endif

	if(ParamIsDefault(createTable))
		 createTable=NumVarOrDefault(NlxFolder+"createTableDefault",0)
	endif
	if(ParamIsDefault(createGraph))
		createGraph=NumVarOrDefault(NlxFolder+"createGraphDefault",0)
	endif
	downSamp=ParamIsDefault(downSamp) ? NumVarOrDefault(NlxFolder+"downsampleNCS",0) : downSamp
	
	String wavesLoaded = ""
	String fileName
	Variable i = 0
	Do
		fileName = IndexedFile($pathName, i, "."+type)
		fileName=RemoveEnding(fileName,"."+type)
		if (strlen(fileName) == 0)
			break										// No more files
		endif
		
		String loadedName = LoadNlxBinaryFile(type,fileName,pathName=pathName,baseName=baseName,createTable=createTable,createGraph=createGraph,downSamp=downSamp)
		Variable err=str2num(loadedName)
		if (!numtype(err) || !strlen(loadedName))
			return err
		endif
		
		if(strlen(loadedName))
			wavesLoaded+=loadedName+";"
		endif
		i += 1
	While(1)
	
	Variable numWavesLoaded
	String name
	
	numWavesLoaded = ItemsInList(wavesLoaded)
	if (numWavesLoaded > 0)
		if (createTable)
			name =  StringFromList(0, wavesLoaded)
			Edit $name
			for(i=1; i<numWavesLoaded; i+=1)
				name =  StringFromList(i, wavesLoaded)
				strswitch(type)
					case "ncs":
						AppendToTable $name // Same table for a new, 1D continuous wave.  
						break
					case "ntt":
						Edit $name // New table for a new, 3D tetrode wave.  
				endswitch
			endfor
		endif

		if (createGraph)
			name =  StringFromList(0, wavesLoaded)
			strswitch(type)
				case "ncs":
					Display $name // Simple graph for contnuous wave.  
					break
				case "ntt":
					CreateNlxViewer($name,type) // Fancy viewer for 3D tetrode wave.  
					break
			endswitch
			for(i=1; i<numWavesLoaded; i+=1)
				name =  StringFromList(i, wavesLoaded)
				strswitch(type)
					case "ncs":
						AppendToGraph $name // Same graph for a new, 1D continuous wave.  
						break
					case "ntt":
						// Nothing for a 3D tetrode wave since we can select it from the viewer.  
						break  
				endswitch
			endfor
		endif
	endif
	
	return 0
End

// ------------------------------ Helper functions. -----------------------------

// Skips num_lines lines worth of data from a file.  
static Function FSkipLines(refNum,num_lines)
	Variable refNum,num_lines
	String dummy
	Variable i
	for(i=0;i<num_lines;i+=1)
		FReadLine refNum,dummy
	endfor
	return num_lines
End